﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using System.Windows;
using System.Windows.Media;
using System.Windows.Shapes;
using Microsoft.Research.DynamicDataDisplay.Charts;
using Microsoft.Research.DynamicDataDisplay.Charts.Legend_items;
using Microsoft.Research.DynamicDataDisplay.DataSources;
using PointInterpolation;


namespace Microsoft.Research.DynamicDataDisplay
{
	/// <summary>
	/// Represents a series of points connected by one polyline.
	/// </summary>
	public class LineGraph : PointsGraphBase
	{
		static LineGraph()
		{
			Type thisType = typeof(LineGraph);

			NewLegend.DescriptionProperty.OverrideMetadata(thisType, new FrameworkPropertyMetadata("LineGraph"));
			NewLegend.LegendItemsBuilderProperty.OverrideMetadata(thisType, new FrameworkPropertyMetadata(new LegendItemsBuilder(DefaultLegendItemsBuilder)));
		}

		private static IEnumerable<FrameworkElement> DefaultLegendItemsBuilder(IPlotterElement plotterElement)
		{
			LineGraph lineGraph = (LineGraph)plotterElement;

			Line line = new Line { X1 = 0, Y1 = 10, X2 = 20, Y2 = 0, Stretch = Stretch.Fill, DataContext = lineGraph };
			line.SetBinding(Line.StrokeProperty, "Stroke");
			line.SetBinding(Line.StrokeThicknessProperty, "StrokeThickness");
			NewLegend.SetVisualContent(lineGraph, line);

			var legendItem = LegendItemsHelper.BuildDefaultLegendItem(lineGraph);
			yield return legendItem;
		}

		private readonly FilterCollection filters = new FilterCollection();

		/// <summary>
		/// Initializes a new instance of the <see cref="LineGraph"/> class.
		/// </summary>
		public LineGraph()
		{
			Legend.SetVisibleInLegend(this, true);
			ManualTranslate = true;

			filters.CollectionChanged += filters_CollectionChanged;
		}

		private void filters_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			filteredPoints = null;
			Update();
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="LineGraph"/> class.
		/// </summary>
		/// <param name="pointSource">The point source.</param>
		public LineGraph(IPointDataSource pointSource)
			: this()
		{
			DataSource = pointSource;
		}

		protected override Description CreateDefaultDescription()
		{
			return new PenDescription();
		}

		/// <summary>Provides access to filters collection</summary>
		public FilterCollection Filters
		{
			get { return filters; }
		}

		#region Pen

		/// <summary>
		/// Gets or sets the brush, using which polyline is plotted.
		/// </summary>
		/// <value>The line brush.</value>
		public Brush Stroke
		{
			get { return LinePen.Brush; }
			set
			{
				if (LinePen.Brush != value)
				{
					if (!LinePen.IsSealed)
					{
						LinePen.Brush = value;
						InvalidateVisual();
					}
					else
					{
						Pen pen = LinePen.Clone();
						pen.Brush = value;
						LinePen = pen;
					}

					RaisePropertyChanged("Stroke");
				}
			}
		}

		/// <summary>
		/// Gets or sets the line thickness.
		/// </summary>
		/// <value>The line thickness.</value>
		public double StrokeThickness
		{
			get { return LinePen.Thickness; }
			set
			{
				if (LinePen.Thickness != value)
				{
					if (!LinePen.IsSealed)
					{
						LinePen.Thickness = value; InvalidateVisual();
					}
					else
					{
						Pen pen = LinePen.Clone();
						pen.Thickness = value;
						LinePen = pen;
					}

					RaisePropertyChanged("StrokeThickness");
				}
			}
		}

		/// <summary>
		/// Gets or sets the line pen.
		/// </summary>
		/// <value>The line pen.</value>
		[NotNull]
		public Pen LinePen
		{
			get { return (Pen)GetValue(LinePenProperty); }
			set { SetValue(LinePenProperty, value); }
		}

		public static readonly DependencyProperty LinePenProperty =
			DependencyProperty.Register(
			"LinePen",
			typeof(Pen),
			typeof(LineGraph),
			new FrameworkPropertyMetadata(
				new Pen(Brushes.Blue, 1),
				FrameworkPropertyMetadataOptions.AffectsRender
				),
			OnValidatePen);

		private static bool OnValidatePen(object value)
		{
			return value != null;
		}

		#endregion

		protected override void OnOutputChanged(Rect newRect, Rect oldRect)
		{
			filteredPoints = null;
			base.OnOutputChanged(newRect, oldRect);
		}

		protected override void OnDataChanged()
		{
			filteredPoints = null;
			base.OnDataChanged();
		}

		protected override void OnDataSourceChanged(DependencyPropertyChangedEventArgs args)
		{
			filteredPoints = null;
			base.OnDataSourceChanged(args);
		}

		protected override void OnVisibleChanged(DataRect newRect, DataRect oldRect)
		{
			if (newRect.Size != oldRect.Size)
			{
				filteredPoints = null;
			}

			base.OnVisibleChanged(newRect, oldRect);
		}

		private FakePointList filteredPoints;
		protected FakePointList FilteredPoints
		{
			get { return filteredPoints; }
			set { filteredPoints = value; }
		}

		protected override void UpdateCore()
		{
			if (DataSource == null) return;
			if (Plotter == null) return;

			Rect output = Viewport.Output;
			var transform = GetTransform();

			if (filteredPoints == null || !(transform.DataTransform is IdentityTransform))
			{
				IEnumerable<Point> points = GetPoints();

				var bounds = BoundsHelper.GetViewportBounds(points, transform.DataTransform);
				Viewport2D.SetContentBounds(this, bounds);

				// getting new value of transform as it could change after calculating and setting content bounds.
				transform = GetTransform();
				List<Point> transformedPoints = transform.DataToScreenAsList(points);

				// Analysis and filtering of unnecessary points
				filteredPoints = new FakePointList(FilterPoints(transformedPoints),
					output.Left, output.Right);

				if (ProvideVisiblePoints)
				{
					List<Point> viewportPointsList = new List<Point>(transformedPoints.Count);
					if (transform.DataTransform is IdentityTransform)
					{
						viewportPointsList.AddRange(points);
					}
					else
					{
						var viewportPoints = points.DataToViewport(transform.DataTransform);
						viewportPointsList.AddRange(viewportPoints);
					}

					SetVisiblePoints(this, new ReadOnlyCollection<Point>(viewportPointsList));
				}

				Offset = new Vector();
			}
			else
			{
				double left = output.Left;
				double right = output.Right;
				double shift = Offset.X;
				left -= shift;
				right -= shift;

				filteredPoints.SetXBorders(left, right);
			}
		}

		StreamGeometry streamGeometry = new StreamGeometry();
		protected override void OnRenderCore(DrawingContext dc, RenderState state)
		{
            // It is possible that filteredPoints is null when the line is first rendered.
		    if (this.DataSource == null || 
                this.filteredPoints == null ||
                !this.filteredPoints.HasPoints)
		    {
		        return;
		    }

		    var points = WillFallFoulOfWpfClipRectIssue(this.filteredPoints)
		                     ? this.CalculateInterpolationPoints(this.filteredPoints)
		                     : this.filteredPoints;
			    
		    using (StreamGeometryContext context = this.streamGeometry.Open())
		    {
		        context.BeginFigure(points.StartPoint, false, false);
		        context.PolyLineTo(points, true, this.smoothLinesJoin);
		    }

		    Brush brush = null;
		    Pen pen = this.LinePen;

		    bool isTranslated = this.IsTranslated;
		    if (isTranslated)
		    {
		        dc.PushTransform(new TranslateTransform(this.Offset.X, this.Offset.Y));
		    }

		    dc.DrawGeometry(brush, pen, this.streamGeometry);
				
		    if (isTranslated)
		    {
		        dc.Pop();
		    }

#if __DEBUG
				FormattedText text = new FormattedText(points.Count.ToString(),
					CultureInfo.InvariantCulture, FlowDirection.LeftToRight,
					new Typeface("Arial"), 12, Brushes.Black);
				dc.DrawText(text, Viewport.Output.GetCenter());
#endif
		}

        /// <summary>
        /// If its been determined that a line being drawn will fail due to issues
        /// (or limitations) with clipping in WPF, then this method will calculate
        /// points at the edges of the output region.
        /// </summary>
        /// <param name="fakePoints">
        /// The original points from which the interpolated points will be 
        /// calculated.
        /// </param>
        /// <returns>A FakePointList holding the interpolated points.</returns>
        private FakePointList CalculateInterpolationPoints(FakePointList fakePoints)
        {
            var viewport = Viewport.Output;
            viewport.Offset(-Offset.X, -Offset.Y);

            var output = LineChartInterpolator.CalculateInterpolationPoints(fakePoints.StartPoint, fakePoints, viewport, this.StrokeThickness / 2);

            return new FakePointList(output, Viewport.Output.Left, Viewport.Output.Right);
        }

        /// <summary>
        /// Determines if a line being drawn between two points, as identified by 
        /// the start point of the passed in FakePointList and the first and only
        /// enumerable item it returns, will be erroneously clipped by WPF when 
        /// rendererd with a clip rect in place.  Unfortunately at this level, we
        /// have to assume the clip rect is in place.  The magic numbers you see 
        /// were discovered through trial and error.
        /// </summary>
        /// <param name="points">The points to check.</param>
        /// <returns>
        /// True if the horizontal difference between the the first and last 
        /// points is greater than 124985 units, or if the Y extent of all points
        /// is greater than the same threshold.
        /// </returns>
        private static bool WillFallFoulOfWpfClipRectIssue(FakePointList points)
        {
            // .Net 4 seems to bring this down from 249971
            const double MagicNumber = 124985;

            if (points.Any())
            {
                return false;
            }

            var dx = points.Last().X - points.StartPoint.X;
            if (dx < MagicNumber)
            {
                var max = Math.Max(points.StartPoint.Y, points.Max(p => p.Y));
                var min = Math.Min(points.StartPoint.Y, points.Min(p => p.Y));
                return max - min > MagicNumber;
            }
            
            return true;
        }

		private bool filteringEnabled = true;
		public bool FilteringEnabled
		{
			get { return filteringEnabled; }
			set
			{
				if (filteringEnabled != value)
				{
					filteringEnabled = value;
					filteredPoints = null;
					Update();
				}
			}
		}

		private bool smoothLinesJoin = true;
		public bool SmoothLinesJoin
		{
			get { return smoothLinesJoin; }
			set
			{
				smoothLinesJoin = value;
				Update();
			}
		}

		private List<Point> FilterPoints(List<Point> points)
		{
			if (!filteringEnabled)
				return points;

			var filteredPoints = filters.Filter(points, Viewport.Output);

			return filteredPoints;
		}
	}
}
