﻿//-----------------------------------------------------------------------
// <copyright company="Omron Corporation">
// Copyright © Omron Corporation 2014. All Rights Reserved
// </copyright>
//
// COMPANY CONFIDENTIAL.
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;

namespace PointInterpolation
{
    /// <summary>
    /// Calculates a list of points such that no points fall outside the given viewport, while
    /// keeping the shape described by the points equivalent to the original source.
    /// </summary>
    public static class LineChartInterpolator
    {
        /// <summary>
        /// Calculates a list of points such that no points fall outside the given viewport, while
        /// keeping the shape described by the points equivalent to the original source.
        /// </summary>
        /// <param name="startPoint">The first point to start taking calculations from</param>
        /// <param name="points">Ordered list of point describing the shape</param>
        /// <param name="viewport">The viewport bouds withing which to keep points</param>
        /// <param name="lineWidth">The width of the line that will be drawn with the points</param>
        /// <returns>List of interpolated points</returns>
        public static List<Point> CalculateInterpolationPoints(Point startPoint, IEnumerable<Point> points, Rect viewport, double lineWidth)
        {
            var output = new List<Point>();
            var lastPoint = startPoint;
            
            if (IsValidPoint(lastPoint) && !IsOutsideViewPort(lastPoint, viewport))
            {
                output.Add(lastPoint);
            }

            foreach (var point in points.Where(IsValidPoint))
            {
                if (IsOutsideViewPort(lastPoint, viewport))
                {
                    Point intersectionPoint;

                    // Must intersect left, top or bottom of viewport.  It is possible that we get a line outside the
                    // bounds, as the FakePointsList only considers X when deciding relevant points.  Blindly adding
                    // points that haven't been interpolated can lead us back to experiencing the problem we came here 
                    // to solve in the first place.
                    if (Intersection(lastPoint, point, viewport.TopLeft, viewport.BottomLeft, out intersectionPoint) ||
                        IntersectionTopOrBottom(point, lastPoint, viewport, out intersectionPoint, lineWidth))
                    {
                        output.Add(intersectionPoint);
                    }
                }

                if (IsOutsideViewPort(point, viewport))
                {
                    Point intersectionPoint;

                    // Must intersect right, top or bottom of viewport 
                    if (Intersection(lastPoint, point, viewport.TopRight, viewport.BottomRight, out intersectionPoint) ||
                        IntersectionTopOrBottom(lastPoint, point, viewport, out intersectionPoint, lineWidth))
                    {
                        output.Add(intersectionPoint);
                    }
                }
                else
                {
                    output.Add(point);
                }

                lastPoint = point;
            }

            return output;
        }

        private static bool IntersectionTopOrBottom(Point lastPoint, Point point, Rect viewport, out Point result, double lineWidth)
        {
            var top = point.Y <= 0;

            var intersected = top
                ? Intersection(lastPoint, point, viewport.TopLeft, viewport.TopRight, out result)
                : Intersection(lastPoint, point, viewport.BottomLeft, viewport.BottomRight, out result);

            if (intersected)
            {
                result.Y += top ? 0 - lineWidth : lineWidth;
            }

            return intersected;
        }

        private static bool IsValidPoint(Point point)
        {
            return !(double.IsInfinity(point.X) ||
                     double.IsInfinity(point.Y) ||
                     double.IsNaN(point.X) ||
                     double.IsNaN(point.Y));
        }

        private static bool IsOutsideViewPort(Point p, Rect viewport)
        {
            return !viewport.Contains(p);
        }

        private static bool Intersection(Point p1, Point p2, Point p3, Point p4, out Point result)
        {
            double x1 = p1.X, x2 = p2.X, x3 = p3.X, x4 = p4.X;
            double y1 = p1.Y, y2 = p2.Y, y3 = p3.Y, y4 = p4.Y;

            var d = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4);

            if (d != 0 && !double.IsInfinity(d))
            {
                double pre = (x1 * y2 - y1 * x2), 
                       post = (x3 * y4 - y3 * x4);
                var x = (pre * (x3 - x4) - (x1 - x2) * post) / d;
                var y = (pre * (y3 - y4) - (y1 - y2) * post) / d;

                const double Epsilon = 0.00001;

                if (x > (Math.Min(x1, x2) - Epsilon) &&
                    x < (Math.Max(x1, x2) + Epsilon) &&
                    x > (Math.Min(x3, x4) - Epsilon) &&
                    x < (Math.Max(x3, x4) + Epsilon) &&
                    y > (Math.Min(y1, y2) - Epsilon) &&
                    y < (Math.Max(y1, y2) + Epsilon) &&
                    y > (Math.Min(y3, y4) - Epsilon) &&
                    y < (Math.Max(y3, y4) + Epsilon))
                {
                    result = new Point(x, y);
                    return true;
                }
            }
            
            result = default(Point);
            return false;
        }
    }
}
