﻿using System.Windows;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace PointInterpolation.Tests
{
    /// <summary>
    /// This tests the line interpolation class.  Given a start and series of 
    /// points, along with a viewport, that class should generate a list of points
    /// that when drawn from one to the next, do not go outside the bounds of the 
    /// viewport but which does not substantially alter the shape of what would be
    /// drawn.
    ///</summary>
    [TestClass]
    public class LineChartInterpolatorTest
    {
        private readonly Rect viewport30X30 = new Rect(0, 0, 30, 30);
        private readonly Rect viewport40X30 = new Rect(0, 0, 40, 30);
        private readonly Rect viewport50X30 = new Rect(0, 0, 50, 30);
        private readonly Rect viewport60X30 = new Rect(0, 0, 60, 30);

        #region Simple Single Line Tests
        /// <summary>
        /// Single line, crossing left edge of view port
        ///   +----+
        ///   | b  |
        /// a |    |
        ///   +----+
        /// </summary>
        [TestMethod]
        public void SingleLineCrossingLeft()
        {
            var start = new Point(-10, 20);
            var points = new[] {new Point(10, 10)};
            var expected = new[] { new Point(0, 15), new Point(10, 10) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Single line crossing bottom edge of view port
        /// +----+
        /// | b  |
        /// +----+
        ///  a
        /// </summary>
        [TestMethod]
        public void SingleLineCrossingBottom()
        {
            var start = new Point(10, 40);
            var points = new[] { new Point(20, 20) };
            var expected = new[] { new Point(15, 32), new Point(20, 20) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Single line crossing top of view port
        /// a
        /// +----+
        /// | b  |
        /// +----+
        /// </summary>
        [TestMethod]
        public void SingleLineCrossingTop()
        {
            var start = new Point(10, -20);
            var points = new[] { new Point(20, 20) };
            var expected = new[] { new Point(15, -2), new Point(20, 20) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Single line crossing right of box
        /// +----+
        /// | a  |
        /// |    | b
        /// +----+
        /// </summary>
        [TestMethod]
        public void SingleLineCrossingRight()
        {
            var start = new Point(20, 10);
            var points = new[] { new Point(40, 20) };
            var expected = new[] { new Point(20, 10), new Point(30, 15) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Single line crossing viewport horizontally
        ///  +----+
        /// a|    |
        ///  |    |b
        ///  +----+ 
        /// </summary>
        [TestMethod]
        public void SingleLineCrossingViewPortHorizontally()
        {
            var start = new Point(-10, 10);
            var points = new[] {new Point(40, 20)};
            var expected = new[] {new Point(0, 12), new Point(30, 18)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Single line crossing viewport vertically
        ///   a
        ///  +----+
        ///  |    |
        ///  +----+
        ///      b 
        /// </summary>
        [TestMethod]
        public void SingleLineCrossingViewPortVertically()
        {
            var start = new Point(10, -10);
            var points = new[] { new Point(20, 40) };
            var expected = new[] { new Point(12, -3), new Point(18, 33) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 3);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        ///  +----+
        /// a|    |
        /// b|    |
        ///  +----+
        /// </summary>
        [TestMethod]
        public void SingleLineDoesNotCrossViewPort()
        {
            var start = new Point(-10, 10);
            var points = new[] {new Point(-10, 20)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            Assert.IsTrue(actual.Count == 0);
        }

        #endregion

        #region Sitting on the view port edges tests

        [TestMethod]
        public void LeftEdge()
        {
            var a = new Point(0, 10);
            var b = new Point(10, 10);
            var points = new[] {b};
            var expected = new[] {a, b};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void TopEdge()
        {
            var a = new Point(10, 0);
            var b = new Point(10, 10);
            var points = new[] { b };
            var expected = new[] { a, b };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void RightEdge()
        {
            var a = new Point(30, 10);
            var b = new Point(10, 10);
            var points = new[] { b };
            var expected = new[] { a, b };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void BottomEdge()
        {
            var a = new Point(10, 30);
            var b = new Point(10, 10);
            var points = new[] { b };
            var expected = new[] { a, b };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        #endregion

        #region Far more exciting than the single line tests tests.

        /// <summary>
        /// +---+
        /// |   |
        /// |  c|
        /// +---+
        ///  ab
        /// </summary>
        [TestMethod]
        public void FirstSegmentFallsOutOfBounds()
        {
            var start = new Point(10, 40);
            var points = new[] {new Point(20, 40), new Point(30, 20)};
            var expected = new[] {new Point(25, 32), new Point(30, 20)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport40X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Three lines (4 points), where the first two fall out of bounds
        /// and should not appear in the result.
        /// +---+
        /// |   |
        /// |  d|
        /// +---+
        /// abc
        /// </summary>
        [TestMethod]
        public void MultipleInitialSegmentsFallOutOfBounds()
        {
            var start = new Point(10, 40);
            var points = new[] {new Point(20, 40), new Point(30, 40), new Point(40, 20)};
            var expected = new[] { new Point(35, 32), new Point(40, 20) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport50X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Multiple end segments fall out side the viewport
        /// +----+
        /// |a   |
        /// +----+
        ///   bcd
        /// </summary>
        [TestMethod]
        public void MultipleEndSegmentsFallOutOfBounds()
        {
            var start = new Point(10, 20);
            var points = new[] {new Point(20, 40), new Point(30, 40), new Point(40, 40)};
            var expected = new[] {new Point(10, 20), new Point(15, 32)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport60X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Multiple segments in the middle fall out of bounds
        /// +-----+
        /// |a   e|
        /// +-----+
        ///   bcd
        /// </summary>
        [TestMethod]
        public void MultipleMidSegmentsFallOutOfBounds()
        {
            var start = new Point(10, 20);
            var points = new[] {new Point(20, 40), new Point(30, 40), new Point(40, 40), new Point(50,20)};
            var expected = new[] {new Point(10, 20), new Point(15, 35), new Point(45, 35), new Point(50, 20)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport60X30, 5);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// +---+
        /// | b |
        /// +---+
        ///  a c
        /// </summary>
        [TestMethod]
        public void MultiSegmentsBelowInsideBelow()
        {
            var start = new Point(10, 40);
            var points = new[] {new Point(20, 20), new Point(30, 40)};
            var expected = new[] {new Point(15, 32), new Point(20, 20), new Point(25, 32)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport40X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// +---+
        /// | b |
        /// |  c|
        /// +---+
        ///  a
        /// </summary>
        [TestMethod]
        public void MultiSegmentBelowInsideInside()
        {
            var start = new Point(10, 40);
            var points = new[] {new Point(20, 20), new Point(30, 20)};
            var expected = new[] {new Point(15, 40), new Point(20, 20), new Point(30, 20)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 10);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        ///   b
        /// +---+
        /// |a c|
        /// |   |
        /// +---+
        /// </summary>
        [TestMethod]
        public void MultiSegmentInsideAboveInside()
        {
            var start = new Point(10, 10);
            var points = new[] {new Point(20, -10), new Point(30, 10)};
            var expected = new[] {new Point(10, 10), new Point(15, -2), new Point(25, -2), new Point(30, 10)};

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport40X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Multiple lines cross right across the hieght of the viewport, but none
        /// of the points fall inside the view port.
        ///  a c e
        /// +-----+
        /// |     |
        /// +-----+
        ///   b d
        /// </summary>
        [TestMethod]
        public void MultipleSegementsCrossVertically()
        {
            var start = new Point(10, -10);
            var points = new[] {new Point(20, 40), new Point(30, -10), new Point(40, 40), new Point(50, -10)};
            var expected = new[]
                               {
                                   new Point(12, -3), new Point(18, 33), 
                                   new Point(22, 33), new Point(28, -3),
                                   new Point(32, -3), new Point(38, 33),
                                   new Point(42, 33), new Point(48, -3)
                               };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport60X30, 3);

            CollectionAssert.AreEqual(expected, actual);
        }

        #endregion

        #region Offset viewport tests

        /// <summary>
        /// Offset view ports; negative x offset
        ///   +---+  
        /// {a| {b|
        ///   +---+
        /// </summary>
        [TestMethod]
        public void ViewportOffsetNegativeHorizontalNoInterpolation()
        {
            var a = new Point(-10, 10);
            var b = new Point(10, 10);
            var points = new[] {b};
            var expected = new[] {a, b};
            var view = new Rect(0, 0, 30, 30);
            view.Offset(-20, 0);

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, view, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Offset view ports; negative x offset
        ///    +---+  
        /// {a | {b|
        ///    +---+
        /// </summary>
        [TestMethod]
        public void ViewportOffsetNegativeHorizontalInterpolated()
        {
            var a = new Point(-20, 10);
            var b = new Point(10, 10);
            var points = new[] { b };
            var expected = new[] {new Point(-10, 10), b};
            var view = new Rect(0, 0, 30, 30);
            view.Offset(-10, 0);

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, view, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Offset view ports; negative y offset
        ///   {a        +----+
        /// +----+  ->  | {a |
        /// | {b |      | {b |
        /// +----+      +----+
        /// </summary>
        [TestMethod]
        public void ViewportOffsetNegativeVerticalNoInterpolation()
        {
            var a = new Point(10, -10);
            var b = new Point(10, 10);
            var points = new[] { b };
            var expected = new[] { a, b };
            var view = new Rect(0, 0, 30, 30);
            view.Offset(0, -20);

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, view, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// Offset view ports; negative y offset
        ///   {a          
        ///               {a
        /// +----+  ->  +----+
        /// | {b |      | {b |
        /// +----+      +----+
        /// </summary>
        [TestMethod]
        public void ViewportOffsetNegativeVerticalInterpolated()
        {
            var a = new Point(10, -20);
            var b = new Point(10, 10);
            var points = new[] { b };
            var expected = new[] { new Point(10, -12), b };
            var view = new Rect(0, 0, 30, 30);
            view.Offset(0, -10);

            var actual = LineChartInterpolator.CalculateInterpolationPoints(a, points, view, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        #endregion

        #region Large value tests

        /// <summary>
        /// 100million-unit length line crossing horizontally
        ///      +---+
        /// a    |   |    b
        ///      +---+
        /// </summary>
        [TestMethod]
        public void LongLineCrossesViewPortHorizontally()
        {
            var start = new Point(-50000000, 10);
            var points = new[] { new Point(50000000, 10) };
            var expected = new[] { new Point(0, 10), new Point(30, 10) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        /// <summary>
        /// 100million-unit length line crossing vertically
        ///   a
        /// 
        /// +---+
        /// |   |
        /// +---+
        /// 
        ///   b
        /// </summary>
        [TestMethod]
        public void LongLineCrossesViewPortVertically()
        {
            var start = new Point(10, -50000000);
            var points = new[] { new Point(10, 50000000) };
            var expected = new[] { new Point(10, -2), new Point(10, 32) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            CollectionAssert.AreEqual(expected, actual);
        }

        #endregion

        #region Invalid numbers tests

        [TestMethod]
        public void RidiculouslyLongLineCrossesViewPortHorizontallyNoOutput()
        {
            var start = new Point(double.MinValue, 10);
            var points = new[] { new Point(double.MaxValue, 10) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);
            
            // TODO: Is it reasonable to expect this to return a populated result set?
            Assert.IsTrue(actual.Count == 0);
            //CollectionAssert.AreEqual(expected, actual);
        }

        [TestMethod]
        public void InfinityCheck1()
        {
            var start = new Point(double.NegativeInfinity, 10);
            var points = new[] { new Point(double.PositiveInfinity, 10) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            Assert.IsTrue(actual.Count == 0);
        }

        [TestMethod]
        public void InfinityCheck2()
        {
            var start = new Point(-10, double.NegativeInfinity);
            var points = new[] { new Point(10, double.PositiveInfinity) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            Assert.IsTrue(actual.Count == 0);
        }

        [TestMethod]
        public void NanCheck1()
        {
            var start = new Point(double.NaN, 10);
            var points = new[] { new Point(double.NaN, 10) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            Assert.IsTrue(actual.Count == 0);
        }

        [TestMethod]
        public void NanCheck2()
        {
            var start = new Point(-10, double.NaN);
            var points = new[] { new Point(10, double.NaN) };

            var actual = LineChartInterpolator.CalculateInterpolationPoints(start, points, viewport30X30, 2);

            Assert.IsTrue(actual.Count == 0);
        }

        #endregion
    }
}
